////////////////////////////////////////////////////////////////
// Advanced Encryption Standard (AES) Implementierung mit
// FreePascal nach
// "Federal Information Processing Standards Publication 197",
// November 26, 2001;
// Autor: Henner Heck, 2005 (HennerHeck@web.de)
////////////////////////////////////////////////////////////////


unit uaes;

{$mode objfpc}{$H+}

{$INLINE ON} // Inline-Kompilierung aktiv
{$DEFINE AesSpeedup} // Benutzung von Wertetabellen anstelle von aesModProd

// Hinweis: Obwohl die Eigenschaft Size von TStream mit Int64 arbeitet,
// sind mit FreePascal 2.0 nur Dateien bis 2GB zuverlssig verarbeitbar,
// denn die dahinterstehende SEEK Prozedur arbeitet mit LongInt

////////////////////////////////////////////////////////////////
// Umgehung der 2GB-Grenze in Win32
// Read und Write von TStream kommen offenbar mit >2GB zurecht
// Getestet unter Windows XP
{$IFDEF WIN32}
  {$DEFINE FILESIZE64}
{$ENDIF}
////////////////////////////////////////////////////////////////

interface

uses
  Classes, SysUtils, umisc;


type TAESKeySize   = (ks128,ks192,ks256);
type TAESBlockSize = (bs128,bs192,bs256);
type TAESState     = Array[0..3,0..3] of Byte;
type TAESBuffer    = Array[0..SizeOf(TAESState)-1] of Byte;
type TAESHHHeader  = Record
//Dateiheader,Informationen ber Schlssellnge und Zusatz-Bytes
     KS:Byte;
     GarbageCount:Byte;
     K0:Byte;
     K1:Byte;
     K2:Byte;
     K3:Byte;
     K4:Byte;
     K5:Byte;
     end;

type TAES = class(TObject)
  protected
    Initiated : Boolean;
    Nb,Nk,Nr : Integer;
    CurrentKeySize:TAESKeySize;
    CipherKey,W : Array of LongWord;
    function ValidHeader(const H:TAESHHHeader):Boolean;
    function MakeHeader(var H:TAESHHHeader; GaCount:Byte; KSize:TAESKeySize):Boolean;
    function KeySize(const KS : TAESKeySize) : Byte;
    function BlockSize(const BS : TAESBlockSize) : Byte;
    function NumberOfRounds(const KS : TAESKeySize; const BS : TAESBlockSize) : Integer;
    procedure EncodeState(var s: TAESState); inline;
    procedure DecodeState(var s: TAESState); inline;
    procedure FillStateFromBuffer(var s: TAESState; const b: TAESBuffer); inline;
    procedure FillBufferFromState(const s: TAESState; var b: TAESBuffer); inline;
    {$IFDEF FILESIZE64}
    procedure EnDeCodeStream(const fS,fD: TStream; const Encode:Boolean; const SPI: TSetProgressInfo; const CA:TSetBool; const MOut:TMesssageOut;const Size64:Int64);
    {$ELSE}
    procedure EnDeCodeStream(const fS,fD: TStream; const Encode:Boolean; const SPI: TSetProgressInfo; const CA:TSetBool; const MOut:TMesssageOut);
    {$ENDIF}
    // MixColumns
    {$IFNDEF AESSPEEDUP}
    function  PolyMul(const P1,P2:Byte):Word; inline;
    procedure PolyDivMod(Dividend,Divisor:Word;var Quotient,Rest:Word); inline;
    function  aesPolyMul(const P1,P2:Byte):Byte; inline;
    procedure aesModProd(const a0,a1,a2,a3,b0,b1,b2,b3:byte; var d0,d1,d2,d3:byte); inline;
    {$ENDIF}
    procedure MixColumn(var s0,s1,s2,s3:byte); inline;
    procedure InvMixColumn(var s0,s1,s2,s3:byte); inline;
    procedure MixColumns(var s:TAESState); inline;
    procedure InvMixColumns(var s:TAESState); inline;
    // SubByte
    function  SubByte(const InByte:Byte): Byte; inline;
    function  InvSubByte(const InByte:Byte): Byte; inline;
    procedure SubBytes(var s:TAESState); inline;
    procedure InvSubBytes(var s:TAESState); inline;
    // ShiftRows
    procedure RotateRow(var s0,s1,s2,s3:byte;const value:Byte; const left:Boolean); inline;
    procedure ShiftRow(var s0,s1,s2,s3:byte;const value:Byte); inline;
    procedure InvShiftRow(var s0,s1,s2,s3:byte;const value:Byte); inline;
    procedure ShiftRows(var s:TAESState); inline;
    procedure InvShiftRows(var s:TAESState); inline;
    // KeyExpansion
    function  SubWord(const InWord:LongWord):LongWord; inline;
    function  RotWord(const InWord:LongWord):LongWord; inline;
    procedure KeyExpansion;
    // AddRoundKey
    procedure AddKeyToCol(var s0,s1,s2,s3:byte; const Key:LongWord); inline;
    procedure AddRoundKey(var s:TAESState; const round:Integer); inline;
  public
    constructor Create(const Key: String; const KS : TAESKeySize);
    procedure   SetNewKey(const Key:String; const KS:TAESKeySize);
    procedure   EnDecodeFile(const Source,Dest:String; const Encode:Boolean; const SPI:TSetProgressInfo; const CA:TSetBool; const MOut:TMesssageOut);
    function    EnDecodeString(const S:String; const Encode:Boolean; const SPI:TSetProgressInfo; const CA:TSetBool; const MOut:TMesssageOut):String;
 end;


implementation

uses uMain;

{$i aes.inc}

constructor TAES.Create(const Key:String; const KS:TAESKeySize);
begin
  inherited Create;
  SetNewKey(Key,KS);
end;

procedure TAES.SetNewKey(const Key:String; const KS:TAESKeySize);
const BS = bs128; // Blockgrsse ist immer 128 Bit -> Nb=4
var KSize,i : Integer;
begin
  Initiated := FALSE;
  SetLength(CipherKey,0);            //Schlssel initialisieren
  Nb:=0; Nk:=0;Nr:=0;                // Werte initialisieren
  KSize := KeySize(KS);              //erwartete Schlsselgrsse in 32Bit-Worten
  if (Length(Key) >= (KSize*4)) then
    begin
      SetLength(CipherKey,KSize);
      Move(Key[1],CipherKey[0],KSize*4); //Inhalt von Key in CipherKey kopieren
      i := 0;
      if LittleEndian then // Bytereihenfolge umkehren (Little Endian)
        begin
          while i < Length(CipherKey) do
            begin
              TurnWord(CipherKey[i]);
              Inc(i);
            end;
        end;
      Nb := BlockSize(BS);                  // Nb festlegen
      Nk := KSize;                          // Nk festlegen
      Nr := NumberOfRounds(KS,BS);          // Nr festlegen
      KeyExpansion;
      CurrentKeySize := KS;
      Initiated := TRUE;
    end;
end;

function TAES.ValidHeader(const H:TAESHHHeader):Boolean;
begin
  Result :=            (H.K0 = $48);
  Result := Result AND (H.K1 = $48);
  Result := Result AND (H.K2 = $48);
  Result := Result AND (H.K3 = $41);
  Result := Result AND (H.K4 = $45);
  Result := Result AND (H.K5 = $53);
  Result := Result AND (H.GarbageCount in [0..15]);
  Result := Result AND (H.KS in [4,6,8]);
end;

function TAES.MakeHeader(var H:TAESHHHeader; GaCount:Byte; KSize:TAESKeySize):Boolean;
begin
  H.K0 := $48;
  H.K1 := $48;
  H.K2 := $48;
  H.K3 := $41;
  H.K4 := $45;
  H.K5 := $53;
  H.GarbageCount := GaCount;
  H.KS := KeySize(KSize);
  Result := ValidHeader(H);
end;

{$IFDEF FILESIZE64}
function FileSize(FileName: String):Int64;
var SR: TSearchRec;
begin
  Result := 0;
  if FindFirst(FileName,faAnyFile,SR) = 0 then
    Result :=  Int64(SR.FindData.nFileSizeHigh) shl Int64(32)
              +Int64(SR.FindData.nFileSizeLow);
  FindClose(SR);
end;
{$ENDIF}

procedure TAES.EnDecodeFile(const Source,Dest:String; const Encode:Boolean; const SPI:TSetProgressInfo; const CA:TSetBool; const MOut:TMesssageOut);
var S,D    : String;
    fS,fD  : TFileStream;
    Weitermachen: Boolean;
begin
  try
    Weitermachen := TRUE;
    S := Trim(Source);
    D := Trim(Dest);
    if S<>D then
      begin
        if MOut<>NIL then
          begin
            MOut('');
            MOut('Quelldatei "' +S +'"');
            MOut('Zieldatei  "' +D +'"');
          end;
        fS := TFileStream.Create(S,fmOpenRead OR fmShareDenyWrite);
        try
          if FileExists(D) then
            begin
              if MOut<>NIL then MOut('Zieldatei existiert bereits');
              DeleteFile(D);
              if MOut<>NIL then MOut('Zieldatei wurde gelscht');
            end;
          fD := TFileStream.Create(D,fmCreate OR fmShareDenyWrite);
          try
            {$IFDEF FILESIZE64}
            EnDeCodeStream(fS,fD,Encode,SPI,CA,MOut,FileSize(S));
            {$ELSE}
            EnDeCodeStream(fS,fD,Encode,SPI,CA,MOut);
            {$ENDIF}
          finally
            fD.Free;
          end;
        finally
          fS.Free;
        end;
      end
    else
      begin
        raise Exception.Create('Die Dateipfade sind identisch!');
      end;
  except
    On E:Exception do
      if MOut<>NIL then MOut(E.Message);
  end;
end;

function TAES.EnDecodeString(const S:String; const Encode:Boolean; const SPI:TSetProgressInfo; const CA:TSetBool; const MOut:TMesssageOut):String;
var fS,fD: TMemoryStream;
begin
  Result := '';
  fS := TMemoryStream.Create;
  CopyStringToStream(S,fS);
  try
    fD := TMemoryStream.Create;
    try
      {$IFDEF FILESIZE64}
      EnDeCodeStream(fS,fD,Encode,SPI,CA,MOut,fS.Size);
      {$ELSE}
      EnDeCodeStream(fS,fD,Encode,SPI,CA,MOut);
      {$ENDIF}
      CopyStreamToString(fD,Result);
    finally
      fD.Free;
    end;
  finally
    fS.Free;
  end;
end;

{$IFDEF FILESIZE64}
procedure TAES.EnDeCodeStream(const fS,fD: TStream; const Encode:Boolean; const SPI: TSetProgressInfo; const CA:TSetBool; const MOut:TMesssageOut;const Size64:Int64);
{$ELSE}
procedure TAES.EnDeCodeStream(const fS,fD: TStream; const Encode:Boolean; const SPI: TSetProgressInfo; const CA:TSetBool; const MOut:TMesssageOut);
{$ENDIF}
const CancelDistance = 3200;
      UpdateDistance = 6400;
var state  : TAESState;
    buffer : TAESBuffer;
    i      : LongInt;
    counter: Int64;
    SPInfo : TSetProgressInfo;
    CAN    : TSetBool;
    MessOut: TMesssageOut;
    Cancel : Boolean;
    Header : TAESHHHeader;
    Garbage: Byte;
    {$IFDEF FILESIZE64}
    CurrPos       : Int64;
    {$ENDIF}
begin
  if Initiated then
    begin
      if ((fS<>NIL) AND (fD<>NIL)) then
        begin
          SPInfo := SPI;
          try if (SPInfo<>NIL) then SPInfo(0); except SPInfo := NIL; end;
          Cancel := FALSE; CAN := CA;
          try if (CAN<>NIL) then CAN(Cancel); except CAN := NIL; end;
          MessOut := MOut;
          try
            if (MessOut<>NIL) then
              begin
                {$IFDEF FILESIZE64}
                MessOut('Routinen fr Dateigren >2GB werden genutzt');
                {$ENDIF}
                if Encode then MessOut('Starte Verschlsselung')
                else           MessOut('Starte Entschlsselung');
              end;
          except MessOut := NIL; end;
          fS.Seek(0,soFromBeginning);
          fD.Seek(0,soFromBeginning);
          Garbage := 0;
          {$IFDEF FILESIZE64}
          CurrPos := 0;
          {$ENDIF}
          if Encode then
            begin
              MakeHeader(Header,Garbage,CurrentKeySize);
              fD.Write(Header,SizeOf(Header));
            end
          else
            begin
              {$IFDEF FILESIZE64}
              if Size64 > SizeOf(Header) then
              {$ELSE}
              if fS.Size > SizeOf(Header) then
              {$ENDIF}
                begin
                  fS.Read(Header,SizeOf(Header));
                  if ValidHeader(Header) then
                    begin
                      Garbage := Header.GarbageCount;
                      {$IFDEF FILESIZE64}
                      CurrPos := CurrPos + Int64(SizeOf(Header));
                      {$ENDIF}
                      if MessOut<>NIL then MessOut('Header gefunden');
                      if MessOut<>NIL then MessOut(IntToStr(Garbage)+' zustzliche Bytes sind vorhanden');
                      if Header.KS <> KeySize(CurrentKeySize) then
                        begin
                          if MessOut<>NIL then MessOut('Schlssellnge laut Header: '+IntToStr((Header.KS)*32) +'Bit, aktuelle Schlssellnge: '+IntToStr(KeySize(CurrentKeySize)*32)+'Bit');
                        end;
                    end
                  else
                    begin
                      fS.Seek(0,soFromBeginning);
                    end;
                end;
            end;
          counter := 0;
          i := fS.Read(buffer[0],SizeOf(buffer));
          while (i = SizeOf(buffer)) AND (NOT Cancel) do
            begin
              Inc(counter);
              {$IFDEF FILESIZE64} CurrPos := CurrPos +Int64(i); {$ENDIF}
              FillStateFromBuffer(state,buffer);
              if Encode then
                begin
                  EncodeState(state);
                  FillBufferFromState(state,buffer);
                  fD.Write(buffer[0],SizeOf(buffer));
                end
              else
                begin
                  DecodeState(state);
                  FillBufferFromState(state,buffer);
                  {$IFDEF FILESIZE64}
                  if ((Size64 -SizeOf(buffer)) < CurrPos) then
                  {$ELSE}
                  if ((fS.Size -SizeOf(buffer)) < fS.Position) then
                  {$ENDIF}
                    fD.Write(buffer[0],SizeOf(buffer)-Garbage)
                  else
                    fD.Write(buffer[0],SizeOf(buffer));
                end;
              if (SPInfo<>NIL) AND ((counter mod UpdateDistance)=0) then
                begin
                  {$IFDEF FILESIZE64}
                  SPInfo(Trunc((CurrPos/Size64)*100));
                  {$ELSE}
                  SPInfo(Trunc((fS.Position/fS.Size)*100));
                  {$ENDIF}
                end;
              if (CAN<>NIL) AND ((counter mod CancelDistance)=0) then
                CAN(Cancel);
              i := fS.Read(buffer[0],SizeOf(buffer));
            end;
          if ((i>0) AND Encode AND (NOT Cancel)) then
            begin
              Garbage := Byte(SizeOf(buffer) -i);
              {$IFDEF FILESIZE64} CurrPos := CurrPos +Int64(i); {$ENDIF}
              Randomize;
              while i < SizeOf(buffer) do
                begin
                  buffer[i] := Byte(Random($FF));
                  Inc(i);
                end;
              FillStateFromBuffer(state,buffer);
              EncodeState(state);
              FillBufferFromState(state,buffer);
              fD.Write(buffer[0],SizeOf(buffer));
              {$IFDEF FILESIZE64}
              if MessOut<>NIL then MessOut(IntToStr(CurrPos)+' Bytes verarbeitet');
              {$ELSE}
              if MessOut<>NIL then MessOut(IntToStr(fS.Position)+' Bytes verarbeitet');
              {$ENDIF}
              if MessOut<>NIL then MessOut(IntToStr(Garbage)+' zustzliche Bytes geschrieben');
              MakeHeader(Header,Garbage,CurrentKeySize);
              fD.Seek(0,soFromBeginning);
              fD.Write(Header,SizeOf(Header));
              if MessOut<>NIL then MessOut('Header geschrieben');
              {$IFDEF FILESIZE64}
              if (SPInfo<>NIL) then SPInfo(Trunc((CurrPos/Size64)*100));
              {$ELSE}
              if (SPInfo<>NIL) then SPInfo(Trunc((fS.Position/fS.Size)*100));
              {$ENDIF}
            end
          else
            begin
              if (NOT Cancel) then
                begin
                  {$IFDEF FILESIZE64}
                  if MessOut<>NIL then MessOut(IntToStr(CurrPos)+' Bytes verarbeitet');
                  {$ELSE}
                  if MessOut<>NIL then MessOut(IntToStr(fS.Position)+' Bytes verarbeitet');
                  {$ENDIF}
                  if Encode AND (MessOut<>NIL) then MessOut('Header geschrieben');
                end;
            end;
          if (MessOut<>NIL) then
            begin
              if Cancel then MessOut('Abbruch durch Benutzer')
              else MessOut('Fertig');
            end;
        end
      else
        begin
          raise Exception.Create('Quelle und/oder Ziel sind nicht verfgbar!');
        end;
    end
  else
    begin
      raise Exception.Create('Es ist kein Schlssel gesetzt!');
    end;
end;

function TAES.KeySize(const KS : TAESKeySize):Byte;
//Schlssellnge in 32Bit-Worten
begin
  if (KS = ks128) then
    Result := 4
  else if (KS = ks192) then
    Result := 6
  else
    Result := 8;
end;

function TAES.BlockSize(const BS : TAESBlockSize):Byte;
//Blockgrsse in 32Bit-Worten, wird fr AES nicht bentigt
begin
  if (BS = bs128) then
    Result := 4
  else if (BS = bs192) then
    Result := 6
  else
    Result := 8;
end;

function TAES.NumberOfRounds(const KS:TAESKeySize; const BS:TAESBlockSize):Integer;
//Rundenanzahl
begin
  if (KS = ks256) or (BS = bs256) then
    Result := 14
  else if (KS = ks192) or (BS = bs192) then
    Result := 12
  else
    Result := 10;
end;

procedure TAES.EncodeState(var s:TAESState); inline;
var round : Integer;
begin
  AddRoundKey(s,0);
  for round := 1 to (Nr-1) do
    begin
      SubBytes(s);
      ShiftRows(s);
      MixColumns(s);
      AddRoundKey(s,round);
    end;
  SubBytes(s);
  ShiftRows(s);
  AddRoundKey(s,Nr);
end;

procedure TAES.DecodeState(var s:TAESState); inline;
var round : Integer;
begin
  AddRoundKey(s,Nr);
  for round := (Nr-1) downto 1 do
    begin
      InvShiftRows(s);
      InvSubBytes(s);
      AddRoundKey(s,round);
      InvMixColumns(s);
    end;
  InvShiftRows(s);
  InvSubBytes(s);
  AddRoundKey(s,0);
end;

procedure TAES.FillStateFromBuffer(var s:TAESState; const b:TAESBuffer); inline;
var i,j,k: Integer;
begin
  k := 0;
  for i:=0 to 3 do
    for j:=0 to 3 do
      begin
        s[j,i] := b[k];
        Inc(k);
      end;
end;

procedure TAES.FillBufferFromState(const s:TAESState; var b:TAESBuffer); inline;
var i,j,k: Integer;
begin
  k := 0;
  for i:=0 to 3 do
    for j:=0 to 3 do
      begin
        b[k] := s[j,i];
        Inc(k);
      end;
end;

///////////////////////MixColumns///////////////////////////////

{$IFDEF AESSPEEDUP}

procedure TAES.MixColumn(var s0,s1,s2,s3:byte); inline;
var a0,a1,a2,a3:Byte;
begin
  a0 := PolyMul_02[s0] XOR PolyMul_03[s1] XOR PolyMul_01[s2] XOR PolyMul_01[s3];
  a1 := PolyMul_01[s0] XOR PolyMul_02[s1] XOR PolyMul_03[s2] XOR PolyMul_01[s3];
  a2 := PolyMul_01[s0] XOR PolyMul_01[s1] XOR PolyMul_02[s2] XOR PolyMul_03[s3];
  a3 := PolyMul_03[s0] XOR PolyMul_01[s1] XOR PolyMul_01[s2] XOR PolyMul_02[s3];
  s0 := a0; s1 := a1; s2 := a2; s3 := a3;
end;

procedure TAES.InvMixColumn(var s0,s1,s2,s3:byte); inline;
var a0,a1,a2,a3:Byte;
begin
  a0 := PolyMul_0e[s0] XOR PolyMul_0b[s1] XOR PolyMul_0d[s2] XOR PolyMul_09[s3];
  a1 := PolyMul_09[s0] XOR PolyMul_0e[s1] XOR PolyMul_0b[s2] XOR PolyMul_0d[s3];
  a2 := PolyMul_0d[s0] XOR PolyMul_09[s1] XOR PolyMul_0e[s2] XOR PolyMul_0b[s3];
  a3 := PolyMul_0b[s0] XOR PolyMul_0d[s1] XOR PolyMul_09[s2] XOR PolyMul_0e[s3];
  s0 := a0; s1 := a1; s2 := a2; s3 := a3;
end;

{$ELSE}

function TAES.PolyMul(const P1,P2:Byte):Word; inline;
// Modulo 2 (XOR) Multiplikation zweier Polynome 7. Grades
var P1_temp,i:Word;
begin
  Result := 0;
  P1_temp  := P1;
  for i := 0 to ((SizeOf(P2)*8)-1) do
    begin
      if ((1 shl i) AND P2) > 0 then
        begin
          Result := Result XOR (P1_temp shl i);
        end;
    end;
end;

procedure TAES.PolyDivMod(Dividend,Divisor:Word;var Quotient,Rest:Word); inline;
// Modulo 2 (XOR) Division zweier Polynome 7. Grades mit Restbildung
var testval,zaehler,i : Word;
begin
  Quotient := 0;
  Rest     := 0;
  if Divisor <> 0 then
    begin
      testval := 1;
      testval  := testval shl ((SizeOf(testval)*8)-1);
      zaehler := 0;
      while (Divisor AND testval) = 0 do
        begin
          Divisor := Divisor shl 1;
          Inc(zaehler);
        end;
      for i := 0 to zaehler do
        begin
          if (Dividend AND (testval shr i)) >0 then
            begin
              Dividend := Dividend XOR Divisor;
              Quotient := Quotient OR 1;
            end;
          Quotient := Quotient shl 1;
          Divisor  := Divisor  shr 1;
        end;
      Rest := Dividend;
    end;
end;

function TAES.aesPolyMul(const P1,P2:Byte):Byte; inline;
// Modulo 2 (XOR) Multiplikation zweier Polynome 7. Grades,
// anschliessendes Modulo mit x^8 +x^4 +x^3 +x +1
// -> Multiplikation im finiten Feld GF(2^8)
const IrrPol8 = $011B; // irreduzibles Polynom 8. Grades: x^8 +x^4 +x^3 +x +1
var Q,R:Word;
begin
  PolyDivMod(PolyMul(P1,P2),IrrPol8,Q,R);
  Result := Byte(R);
end;

procedure TAES.aesModProd(const a0,a1,a2,a3,b0,b1,b2,b3:byte; var d0,d1,d2,d3:byte); inline;
// Multiplikation zweier Polynome 3. Grades a,b modulo (x^4 +1)
begin
  d0 := aesPolyMul(a3,b0) XOR aesPolyMul(a0,b1) XOR aesPolyMul(a1,b2) XOR aesPolyMul(a2,b3);
  d1 := aesPolyMul(a2,b0) XOR aesPolyMul(a3,b1) XOR aesPolyMul(a0,b2) XOR aesPolyMul(a1,b3);
  d2 := aesPolyMul(a1,b0) XOR aesPolyMul(a2,b1) XOR aesPolyMul(a3,b2) XOR aesPolyMul(a0,b3);
  d3 := aesPolyMul(a0,b0) XOR aesPolyMul(a1,b1) XOR aesPolyMul(a2,b2) XOR aesPolyMul(a3,b3);
end;

procedure TAES.MixColumn(var s0,s1,s2,s3:byte); inline;
const a: array[0..3] of byte = ($03,$01,$01,$02);
var s0_,s1_,s2_,s3_:byte;
begin
  aesModProd(a[0],a[1],a[2],a[3],s0,s1,s2,s3,s0_,s1_,s2_,s3_);
  s0 := s0_;
  s1 := s1_;
  s2 := s2_;
  s3 := s3_;
end;

procedure TAES.InvMixColumn(var s0,s1,s2,s3:byte); inline;
const a: array[0..3] of byte = ($0b,$0d,$09,$0e);
var s0_,s1_,s2_,s3_:byte;
begin
  aesModProd(a[0],a[1],a[2],a[3],s0,s1,s2,s3,s0_,s1_,s2_,s3_);
  s0 := s0_;
  s1 := s1_;
  s2 := s2_;
  s3 := s3_;
end;

{$ENDIF}

procedure TAES.MixColumns(var s:TAESState); inline;
var i : Integer;
begin
  for i := 0 to 3 do
    MixColumn(s[0,i],s[1,i],s[2,i],s[3,i]);
end;

procedure TAES.InvMixColumns(var s:TAESState); inline;
var i : Integer;
begin
  for i := 0 to 3 do
    InvMixColumn(s[0,i],s[1,i],s[2,i],s[3,i]);
end;

////////////////////////SubByte/////////////////////////////////

function TAES.SubByte(const InByte:Byte): Byte; inline;
begin
  Result := SBox[InByte];
end;

function TAES.InvSubByte(const InByte:Byte): Byte; inline;
begin
  Result := InvSBox[InByte];
end;

procedure TAES.SubBytes(var s:TAESState); inline;
var i,j:Integer;
begin
  for i:=0 to Nb-1 do
    begin
      for j:=0 to 3 do
        s[i,j] := SubByte(s[i,j]);
    end;
end;

procedure TAES.InvSubBytes(var s:TAESState); inline;
var i,j:Integer;
begin
  for i:=0 to (Nb-1) do
    begin
      for j:=0 to 3 do
        s[i,j] := InvSubByte(s[i,j]);
    end;
end;

////////////////////////ShiftRows///////////////////////////////

procedure TAES.RotateRow(var s0,s1,s2,s3:byte;const value:Byte; const left:Boolean); inline;
//Rotiert 4 Bytes um den Wert value nach links oder rechts
var S:LongWord;
begin
  if value in [1..3] then
    begin
      S := 0;
      S := S OR s0; S := S shl 8;
      S := S OR s1; S := S shl 8;
      S := S OR s2; S := S shl 8;
      S := S OR s3;
      if left then
        S := (S shl (value*8)) OR (S shr ((SizeOf(S)*8) -(value*8)))
      else
        S := (S shr (value*8)) OR (S shl ((SizeOf(S)*8) -(value*8)));
      s3 := S AND $FF; S := S shr 8;
      s2 := S AND $FF; S := S shr 8;
      s1 := S AND $FF; S := S shr 8;
      s0 := S AND $FF;
    end;
end;

procedure TAES.ShiftRow(var s0,s1,s2,s3:byte;const value:Byte); inline;
begin
  RotateRow(s0,s1,s2,s3,value,true);
end;

procedure TAES.InvShiftRow(var s0,s1,s2,s3:byte;const value:Byte); inline;
begin
  RotateRow(s0,s1,s2,s3,value,false);
end;

procedure TAES.ShiftRows(var s:TAESState); inline;
var i : Byte;
begin
  for i := 0 to 3 do
    ShiftRow(s[i,0],s[i,1],s[i,2],s[i,3],i);
end;

procedure TAES.InvShiftRows(var s:TAESState); inline;
var i : Byte;
begin
  for i := 0 to 3 do
    InvShiftRow(s[i,0],s[i,1],s[i,2],s[i,3],i);
end;

////////////////////////Key Expansion///////////////////////////

function TAES.SubWord(const InWord:LongWord):LongWord; inline;
//ersetzt die Bytes eines 32Bit-Wortes durch ihre S-Box Werte
var s: array[0..3] of byte;
    temp:LongWord;
    i:Integer;
begin
  temp := InWord;
  for i:=0 to 3 do
    begin
      s[i] := SubByte(Byte(temp AND $FF));
      temp := temp shr 8;
    end;
  Result := 0;
  for i:=3 downto 1 do
    begin
      Result := Result OR s[i];
      Result := Result shl 8;
    end;
  Result := Result OR s[0];
end;

function TAES.RotWord(const InWord:LongWord):LongWord; inline;
//rotiert ein 32Bit-Wort um 8Bit nach links
begin
  Result := (InWord shl 8) OR (InWord shr ((SizeOf(InWord)*8) -8));
end;

procedure TAES.KeyExpansion;
//Erzeugung der Rundenschlssel
var KeyIndex,NumberOfKeys : Integer;
    Wtemp : LongWord;
begin
  NumberOfKeys := Nb*(Nr+1);
  SetLength(W,NumberOfKeys);
  KeyIndex := 0;
  while KeyIndex < Nk do
    begin
      W[KeyIndex] := CipherKey[KeyIndex];
      Inc(KeyIndex);
    end;
  KeyIndex := Nk;
  while (KeyIndex < NumberOfKeys) do
    begin
      Wtemp := W[KeyIndex -1];
      if ((KeyIndex mod Nk) = 0) then
        Wtemp := SubWord(RotWord(Wtemp)) XOR Rcon[KeyIndex div Nk]
      else if ((Nk > 6) AND ((KeyIndex mod Nk) = 4)) then
        Wtemp := SubWord(Wtemp);
      W[KeyIndex] := W[KeyIndex -Nk] XOR Wtemp;
      Inc(KeyIndex);
    end;
end;

////////////////////////AddRoundKey/////////////////////////////

procedure TAES.AddKeyToCol(var s0,s1,s2,s3:byte; const Key:LongWord); inline;
var S : LongWord;
begin
  S := 0;
  S := S OR s0; S := S shl 8;
  S := S OR s1; S := S shl 8;
  S := S OR s2; S := S shl 8;
  S := S OR s3;
  S := S XOR Key;
  s3 := S AND $FF; S := S shr 8;
  s2 := S AND $FF; S := S shr 8;
  s1 := S AND $FF; S := S shr 8;
  s0 := S AND $FF;
end;

procedure TAES.AddRoundKey(var s:TAESState; const round:Integer); inline;
var i : Integer;
begin
  for i := 0 to (Nb-1) do
    AddKeyToCol(s[0,i],s[1,i],s[2,i],s[3,i],W[(round*Nb)+i]);
end;

////////////////////////////////////////////////////////////////

end.

