////////////////////////////////////////////////////////////////
// SHA-256 Hash
////////////////////////////////////////////////////////////////

unit usha256;

{$mode objfpc}{$H+}
{$inline on}

interface

uses
  Classes, SysUtils, umisc;

  
function  HashStringSHA256(const s:String; const HexOut:Boolean):String; overload;
procedure HashStringSHA256(const s:String; const SHash:TStream; const SPI: TSetProgressInfo); overload;


implementation

{$i sha256.inc}

type THash256         = Array[0..7] of LongWord;
type TMessageBlock    = Array[0..15] of LongWord;
type TMessageSchedule = Array[0..63] of LongWord;

function CH(const x,y,z: LongWord):LongWord; inline;
begin
  Result := (x AND y) XOR ((NOT x) AND z);
end;

function Maj(const x,y,z: LongWord):LongWord; inline;
begin
  Result := (x AND y) XOR (x AND z) XOR (y AND z);
end;

function ROTR(const InWord:LongWord; val:Byte):LongWord; inline;
//rotiert ein 32Bit-Wort um val Bit nach rechts
var tempval: Byte;
begin
  tempval := val mod (SizeOf(InWord)*8);
  Result := (InWord shr tempval) OR (InWord shl ((SizeOf(InWord)*8) -tempval));
end;

function Sigma_upper_256_0(const x: LongWord): LongWord; inline;
begin
  Result := ROTR(x,2) XOR ROTR(x,13) XOR ROTR(x,22);
end;

function Sigma_upper_256_1(const x: LongWord): LongWord; inline;
begin
  Result := ROTR(x,6) XOR ROTR(x,11) XOR ROTR(x,25);
end;

function Sigma_lower_256_0(const x: LongWord): LongWord; inline;
begin
  Result := ROTR(x,7) XOR ROTR(x,18) XOR (x shr 3);
end;

function Sigma_lower_256_1(const x: LongWord): LongWord; inline;
begin
  Result := ROTR(x,17) XOR ROTR(x,19) XOR (x shr 10);
end;

procedure InitHash(var H: THash256);
var i: Integer;
begin
  for i:= 0 to 7 do
    H[i] := Init256[i];
end;

procedure CopyHash(const S: THash256; var D: THash256); inline;
var i: Integer;
begin
  for i:= 0 to 7 do
    D[i] := S[i];
end;

procedure AddHash(const S: THash256; var D: THash256); inline;
var i: Integer;
begin
  for i:= 0 to 7 do
    D[i] := D[i] +S[i];
end;

procedure TurnHash(var H: THash256);
var i: Integer;
begin
  for i:= 0 to High(H) do
    TurnWord(H[i]);
end;

procedure TurnMessageBlock(var M:TMessageBlock); inline;
var i:Integer;
begin
  for i:=0 to High(M) do
    TurnWord(M[i]);
end;

procedure PrepareMessageSchedule(const M:TMessageBlock; var W:TMessageSchedule); inline;
var t:Integer;
begin
  for t:=0 to 15 do
    W[t] := M[t];
  for t:=16 to 63 do
    W[t] := Sigma_lower_256_1(W[t-2]) +W[t-7] +Sigma_lower_256_0(W[t-15]) +W[t-16];
end;

procedure HashMessageBlock(const W:TMessageSchedule; var H:THash256); inline;
var t:Integer;
    T1,T2:LongWord;
begin
  for t:=0 to 63 do
    begin
      T1 := H[7] +Sigma_upper_256_1(H[4]) +Ch(H[4],H[5],H[6]) +k256[t] +W[t];
      T2 := Sigma_upper_256_0(H[0]) +Maj(H[0],H[1],H[2]);
      H[7] := H[6];
      H[6] := H[5];
      H[5] := H[4];
      H[4] := H[3] +T1;
      H[3] := H[2];
      H[2] := H[1];
      H[1] := H[0];
      H[0] := T1 +T2;
    end;
  T1 := 0; T2 := 0;
end;

procedure MakeSHA256Hash(const SMessage,SHash:TStream; const SPInfo: TSetProgressInfo);
// SMessage: Stream mit den zu hashenden Daten
// SHash: nimmt den Hash auf
// SPInfo: Prozedur zur Fortschrittsanzeige in Prozent
const UpdateDistance = 1000;
var LE : Boolean;
    buffer : TMessageBlock;
    bbuffer : Array[0..SizeOf(buffer)-1] of Byte;
    Hash,WorkingVars : THash256;
    W : TMessageSchedule;
    i,j : LongInt;
    SPI: TSetProgressInfo;
    counter : Int64;
    MessageSize : Int64; // Einschrnkung gegenber dem SHA-256 Standard,
                         // es knnen keine Nachrichtenlngen von 2^64 Bit
                         // verarbeitet werden, da Int64 maximal den Wert
                         // 2^63-1 darstellt

// Hinweis: Obwohl die Eigenschaft Size von TStream mit Int64 arbeitet,
// sind mit FreePascal nur Dateien bis 2GB zuverlssig verarbeitbar,
// denn die dahinterstehende SEEK Prozedur arbeitet mit LongInt

begin
  if ((SMessage<>NIL) AND (SHash<>NIL)) then
    begin
      LE := LittleEndian;
      InitHash(Hash);
      SMessage.Seek(0,soFromBeginning);
      SPI := SPInfo;
      try if (SPI<>NIL) then SPI(0); except SPI := NIL; end;
      counter := 0;
      i := SMessage.Read(buffer[0],SizeOf(buffer));
      while i = SizeOf(buffer) do
        begin
          Inc(counter);
          if LE then TurnMessageBlock(buffer);
          PrepareMessageSchedule(buffer,W);
          CopyHash(Hash,WorkingVars);
          HashMessageBlock(W,WorkingVars);
          AddHash(WorkingVars,Hash);
          if (SPI<>NIL) AND ((counter mod UpdateDistance)=0) then
            SPI(Trunc((SMessage.Position/SMessage.Size)*100));
          i := SMessage.Read(buffer[0],SizeOf(buffer));
        end;
      MessageSize := SMessage.Size*8; //Nachrichtenlnge in Bit
      for j:=Low(bbuffer) to High(bbuffer) do bbuffer[j]:=0;
      Move(buffer[0],bbuffer[0],i);
      bbuffer[i] := $80; //1 mal Bit '1' und 7 mal Bit '0' anhngen
      if i>53 then //mehr als 54 Byte des Blocks bereits belegt
        begin
          Move(bbuffer[0],buffer[0],SizeOf(buffer));
          if LE then TurnMessageBlock(buffer);
          PrepareMessageSchedule(buffer,W);
          CopyHash(Hash,WorkingVars);
          HashMessageBlock(W,WorkingVars);
          AddHash(WorkingVars,Hash);
          for j:=Low(bbuffer) to High(bbuffer) do bbuffer[j]:=0;
         end;
      for j:=High(bbuffer) downto (High(bbuffer)-SizeOf(MessageSize)) do
        begin
          bbuffer[j] := Byte(MessageSize AND $FF);
          MessageSize := MessageSize shr (SizeOf(Byte)*8);
        end;
      Move(bbuffer[0],buffer[0],SizeOf(buffer));
      if LE then TurnMessageBlock(buffer);
      PrepareMessageSchedule(buffer,W);
      CopyHash(Hash,WorkingVars);
      HashMessageBlock(W,WorkingVars);
      AddHash(WorkingVars,Hash);
      if LE then TurnHash(Hash);
      SHash.Seek(0,soFromBeginning);
      SHash.Write(Hash[0],SizeOf(Hash));
      if (SPI<>NIL) then SPI(Trunc((SMessage.Position/SMessage.Size)*100));
    end
  else
    begin
      raise Exception.Create('SHA-256: Quelle und/oder Ziel sind nicht verfgbar!');
    end;
end;

function  HashStringSHA256(const s:String; const HexOut:Boolean):String; overload;
var SMessage,SHash : TMemoryStream;
begin
  Result := '';
  SMessage := TMemoryStream.Create;
  try
    SHash := TMemoryStream.Create;
    try
      CopyStringToStream(s,SMessage);
      MakeSHA256Hash(SMessage,SHash,NIL);
      CopyStreamToString(SHash,Result);
      if HexOut then Result := StringToHex(Result);
    finally
      SHash.Free;
    end;
  finally
    SMessage.Free;
  end;
end;

procedure HashStringSHA256(const s:String; const SHash:TStream; const SPI: TSetProgressInfo); overload;
var SMessage : TMemoryStream;
begin
  SMessage := TMemoryStream.Create;
  try
    CopyStringToStream(s,SMessage);
    MakeSHA256Hash(SMessage,SHash,SPI);
  finally
    SMessage.Free;
  end;
end;


end.

